/*
 * $Id: fontprefs.c,v 1.5 2007/09/09 17:15:05 piru Exp $
 */
#include "global.h"
#include <libraries/iffparse.h>
#include <graphics/rastport.h>
#include <prefs/font.h>
#include <proto/iffparse.h>
#include <proto/graphics.h>
#include <aros/libcall.h>
#include <aros/asmcall.h>

#include <string.h>

#define DEBUG 0
#include <aros/debug.h>


struct Screen *FindWBScreen(void);
void ReOpenWBScreen(struct Screen *MyScreen);


static
BOOL SameTextAttr(const struct TextAttr *ta1, const struct TextAttr *ta2)
{
	UBYTE style1 = ta1->ta_Style & (FSF_UNDERLINED | FSF_BOLD | FSF_ITALIC | FSF_EXTENDED);
	UBYTE style2 = ta2->ta_Style & (FSF_UNDERLINED | FSF_BOLD | FSF_ITALIC | FSF_EXTENDED);
	UBYTE flags1 = ta1->ta_Flags & (FPF_ROMFONT | FPF_DISKFONT | FPF_REVPATH | FPF_TALLDOT | FPF_WIDEDOT| FPF_PROPORTIONAL);
	UBYTE flags2 = ta2->ta_Flags & (FPF_ROMFONT | FPF_DISKFONT | FPF_REVPATH | FPF_TALLDOT | FPF_WIDEDOT| FPF_PROPORTIONAL);

	D(bug("SameTextAttr: <%s> %lu 0x%lx 0x%lx -- <%s> %lu 0x%lx 0x%lx\n",
	      ta1->ta_Name, ta1->ta_YSize, style1, flags1,
	      ta2->ta_Name, ta2->ta_YSize, style2, flags2));

	return ta1->ta_YSize == ta2->ta_YSize &&
	       style1 == style2 &&
	       flags1 == flags2 &&
	       !strcmp(ta1->ta_Name, ta2->ta_Name);
}


AROS_UFH1(LONG, FontPrefs_Handler,
		  AROS_UFHA(struct IFFHandle *, iff, A2))
{
	AROS_LIBFUNC_INIT

	struct FontPrefs prefs;

	D(bug("FontPrefs_Handler: FontPrefs\n"));

	ReadChunkBytes(iff, &prefs, sizeof(prefs));

	D(bug("FontPrefs_Handler: Type %d Name \"%s\"\n", prefs.fp_Type, prefs.fp_Name));

	if (prefs.fp_Type != FP_SCREENFONT)
	{
		int length = strlen(prefs.fp_Name);
		struct WBFontPrefs *wbp;

		D(bug("FontPrefs_Handler: !FP_SCREENFONT\n"));

		if ((wbp=AllocVec(sizeof(*wbp) + length, MEMF_PUBLIC | MEMF_CLEAR)))
		{
			D(bug("FontPrefs_Handler: wbp 0x%lx\n",wbp));
			wbp->TextAttrPtr = &wbp->TextAttr;
			wbp->DrawMode    = prefs.fp_DrawMode;
			wbp->FrontPen    = prefs.fp_FrontPen;
			wbp->BackPen     = prefs.fp_BackPen;
			wbp->TextAttr    = prefs.fp_TextAttr;
			wbp->TextAttr.ta_Name = wbp->Name;
			CopyMem(prefs.fp_Name, wbp->Name, length + 1);

			D(bug("FontPrefs_Handler: TextAttrPtr 0x%lx\n",wbp->TextAttrPtr));
			D(bug("FontPrefs_Handler: DrawMode %ld\n",wbp->DrawMode));
			D(bug("FontPrefs_Handler: FrontPen %ld\n",wbp->FrontPen));
			D(bug("FontPrefs_Handler: BackPen %ld\n",wbp->BackPen));

			if ((wbp->TextFont = OpenDiskFont(&wbp->TextAttr)))
			{
				D(bug("FontPrefs_Handler: TextFont 0x%lx\n",wbp->TextFont));

				if (prefs.fp_Type == FP_WBFONT)
				{
					/*
					 * That's the font used for icons.
					 */
					D(bug("FontPrefs_Handler: WBConfig FP_WBFONT\n"));
					wbp = (struct WBFontPrefs *)WBConfig(3, wbp);

					D(bug("FontPrefs_Handler: CloseFont\n"));
					if (wbp)
					{
						CloseFont(wbp->TextFont);
					}
				}
				else if (prefs.fp_Type == FP_SYSFONT)
				{
					/*
					 * FP_SYSFONT is used for WB listers.
					 */
					D(bug("FontPrefs_Handler: WBConfig FP_SYSFONT\n"));

					wbp = (struct WBFontPrefs *)WBConfig(4, wbp);
					if (wbp)
					{
						CloseFont(wbp->TextFont);
					}
				}
				else
				{
					D(bug("FontPrefs_Handler: CloseFont\n"));
					CloseFont(wbp->TextFont);
				}
			}

			D(bug("FontPrefs_Handler: free wbp\n"));
			FreeVec(wbp);
		}
	}

	if (prefs.fp_Type != FP_WBFONT) /* FP_SCREENFONT/FP_SYSFONT */
	{
#pragma pack(2)
		struct IFontPrefs
		{
			struct TextAttr fp_TextAttr;
			UBYTE           fp_Name[32];
			ULONG           fp_xxx;
			WORD            fp_ScrFont;
		} fprefs;
#pragma pack()
		struct TextFont *font;

		D(bug("FontPrefs_Handler: FP_SCREENFONT/FP_SYSFONT\n"));

		memcpy(&fprefs.fp_TextAttr, &prefs.fp_TextAttr, sizeof(struct TextAttr));
		fprefs.fp_TextAttr.ta_Name = fprefs.fp_Name;
		fprefs.fp_ScrFont          = prefs.fp_Type != FP_SYSFONT;
		stccpy(fprefs.fp_Name, prefs.fp_Name, sizeof(fprefs.fp_Name));

		font = OpenDiskFont(&fprefs.fp_TextAttr);
		if (font)
		{
			struct Screen *MyScreen = NULL;

			D(bug("FontPrefs_Handler: SetIPrefs\n"));

			if (prefs.fp_Type == FP_SCREENFONT)
			{
				MyScreen = FindWBScreen();
			}

			SetIPrefs(&fprefs, sizeof(fprefs), IP_FONT);

			if (prefs.fp_Type == FP_SCREENFONT)
			{
				if (MyScreen)
				{
					D(bug("FontPrefs_Handler: Workbench screen found\n"));

					if (!SameTextAttr(&fprefs.fp_TextAttr, MyScreen->Font))
					{
						D(bug("FontPrefs_Handler: screen font changed\n"));
						D(bug("FontPrefs_Handler: Attempt to reopen Workbench screen\n"));
						ReOpenWBScreen(MyScreen);
					}
					else
					{
						D(bug("FontPrefs_Handler: screen font is old..don't change screen\n"));
					}
				}
				else
				{
					D(bug("FontPrefs_Handler: No Workbench screen found\n"));
				}
			}

			D(bug("FontPrefs_Handler: CloseFont\n"));
			CloseFont(font);
		}
	}

	D(bug("FontPrefs_Handler: done\n"));
	return 0;

	AROS_LIBFUNC_EXIT
}

static
const UBYTE options[]=
	"Type/N/K/A,"
	"FrontPen/N/K,"
	"BackPen/N/K,"
	"DrawMode/N/K,"
	"FontName/K,"
	"FontYSize/N/K,"
	"FontStyle/N/K,"
	"FontFlags/N/K,"
	"/F";


struct Args
{
	ULONG	*Type;
	ULONG	*FrontPen;
	ULONG	*BackPen;
	ULONG	*DrawMode;
	STRPTR	FontName;
	ULONG	*FontYSize;
	ULONG	*FontStyle;
	ULONG	*FontFlags;
	STRPTR	FutureArgs;
};

/*
 * New font prefs
 */
void LoadFontPrefs(char *filename)
{
BPTR		MyFile;
struct Args	args;
char strbuf[128];

	D(bug("%s: New Font <%s>\n",__func__,filename));

	if ((MyFile = Open(filename, MODE_OLDFILE)))
	{
		BPTR		OldFile;
		struct RDArgs rda;
		struct RDArgs *rda1;
		char *point;

		memset(&rda,0,sizeof(struct RDArgs));

		OldFile = Input();
		SelectInput(MyFile);
		rda.RDA_Flags = RDAF_NOPROMPT | RDAF_STDIN;

		for (;;)
		{
			memset(&args,0,sizeof(args));
		        if ((rda1 = ReadArgs(options, (LONG *)&args, &rda)))
        		{
				ULONG	Type;

				Type	=	args.Type ? *args.Type : 0;

				D(bug("%s: Type %ld\n",__func__,
						Type));


				if (Type != FP_SCREENFONT)
				{
					int length = args.FontName ? strlen(args.FontName) : 0;
					struct WBFontPrefs *wbp;

					D(bug("%s: !FP_SCREENFONT\n",__func__));

					if ((wbp=AllocVec(sizeof(*wbp) + length, MEMF_PUBLIC | MEMF_CLEAR)))
					{
						D(bug("%s: wbp 0x%lx\n",__func__,wbp));
						wbp->TextAttrPtr = &wbp->TextAttr;
						wbp->DrawMode    = args.DrawMode ? *args.DrawMode : JAM1;
						wbp->FrontPen    = args.FrontPen ? *args.FrontPen : 1;
						wbp->BackPen     = args.BackPen ? *args.BackPen : 0;
						wbp->TextAttr.ta_YSize = args.FontYSize ? *args.FontYSize : 8;
						wbp->TextAttr.ta_Style = args.FontStyle ? *args.FontStyle : 0;
						wbp->TextAttr.ta_Flags = args.FontFlags ? *args.FontFlags : 0;
						wbp->TextAttr.ta_Name = wbp->Name;
						if (args.FontName)
						{
							CopyMem(args.FontName, wbp->Name, length + 1);
						}

						D(bug("%s: Font <%s> YSize %ld Style 0x%lx Flags 0x%lx\n",__func__,
							wbp->TextAttr.ta_Name,
							wbp->TextAttr.ta_YSize,
							wbp->TextAttr.ta_Style,
							wbp->TextAttr.ta_Flags));

						D(bug("%s: TextAttrPtr 0x%lx\n",__func__,wbp->TextAttrPtr));
						D(bug("%s: DrawMode %ld\n",__func__,wbp->DrawMode));
						D(bug("%s: FrontPen %ld\n",__func__,wbp->FrontPen));
						D(bug("%s: BackPen %ld\n",__func__,wbp->BackPen));

						if ((wbp->TextFont = OpenDiskFont(&wbp->TextAttr)))
						{
							D(bug("%s: TextFont 0x%lx\n",__func__,wbp->TextFont));

							if (Type == FP_WBFONT)
							{
								/*
								 * That's the font used for icons.
								 */
								D(bug("%s: WBConfig FP_WBFONT\n",__func__));
								wbp = (struct WBFontPrefs *)WBConfig(3, wbp);

								D(bug("%s: CloseFont\n",__func__));
								if (wbp)
								{
									CloseFont(wbp->TextFont);
								}
							}
							else if (Type == FP_SYSFONT)
							{
								/*
								 * FP_SYSFONT is used for WB listers.
								 */
								D(bug("%s: WBConfig FP_SYSFONT\n",__func__));

								wbp = (struct WBFontPrefs *)WBConfig(4, wbp);
								if (wbp)
								{
									CloseFont(wbp->TextFont);
								}
							}
							else
							{
								D(bug("%s: CloseFont\n",__func__));
								CloseFont(wbp->TextFont);
							}
						}
						else
						{
							D(bug("%s: OpenDiskFont failed\n",__func__));
						}

						D(bug("%s: free wbp\n",__func__));
						FreeVec(wbp);
					}
				}


				if (Type != FP_WBFONT) /* FP_SCREENFONT/FP_SYSFONT */
				{
					#pragma pack(2)
					struct IFontPrefs
					{
						struct TextAttr fp_TextAttr;
						UBYTE           fp_Name[32];
						ULONG           fp_xxx;
						WORD            fp_ScrFont;
					} fprefs;
					#pragma pack()
					struct TextFont *font;

					D(bug("%s: FP_SCREENFONT/FP_SYSFONT\n",__func__));

					fprefs.fp_TextAttr.ta_YSize = args.FontYSize ? *args.FontYSize : 8;
					fprefs.fp_TextAttr.ta_Style = args.FontStyle ? *args.FontStyle : 0;
					fprefs.fp_TextAttr.ta_Flags = args.FontFlags ? *args.FontFlags : 0;
					fprefs.fp_TextAttr.ta_Name = fprefs.fp_Name;
					if (args.FontName)
					{
						stccpy(fprefs.fp_Name, args.FontName, sizeof(fprefs.fp_Name));
					}

					fprefs.fp_ScrFont          = Type != FP_SYSFONT;

					D(bug("%s: Font <%s> YSize %ld Style 0x%lx Flags 0x%lx ScrFont %ld\n",__func__,
						fprefs.fp_TextAttr.ta_Name,
						fprefs.fp_TextAttr.ta_YSize,
						fprefs.fp_TextAttr.ta_Style,
						fprefs.fp_TextAttr.ta_Flags,
						fprefs.fp_ScrFont));

					font = OpenDiskFont(&fprefs.fp_TextAttr);
					if (font)
					{
						struct Screen *MyScreen = NULL;

						D(bug("%s: SetIPrefs\n",__func__));

						if (Type == FP_SCREENFONT)
						{
							MyScreen = FindWBScreen();
						}

						SetIPrefs(&fprefs, sizeof(fprefs), IP_FONT);

						if (Type == FP_SCREENFONT)
						{
							if (MyScreen)
							{
								D(bug("%s: Workbench screen found\n",__func__));

								if (!SameTextAttr(&fprefs.fp_TextAttr, MyScreen->Font))
								{
									D(bug("%s: screen font changed\n",__func__));
									D(bug("%s: Attempt to reopen Workbench screen\n",__func__));
									ReOpenWBScreen(MyScreen);
								}
								else
								{
									D(bug("%s: screen font is old..don't change screen\n",__func__));
								}
							}
							else
							{
								D(bug("%s: No Workbench screen found\n",__func__));
							}
						}

						D(bug("%s: CloseFont\n",__func__));
						CloseFont(font);
					}
					else
					{
						D(bug("%s: OpenDiskFont failed\n",__func__));
					}
				}

				FreeArgs(rda1);
			}
			else
			{
				D(bug("%s: End IoErr %ld\n",__func__,IoErr()));
				break;
			}
		}



		SelectInput(OldFile);
		Close(MyFile);
	}
	else
	{
		D(bug("%s: no new font prefs found\n",__func__));
	}
	D(bug("%s: done\n",__func__));
}
