#include <exec/memory.h>
#include <exec/semaphores.h>
#include <exec/tasks.h>
#include <exec/execbase.h>
#include <exec/rawfmt.h>
#include <dos/dos.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/sysdebug.h>
#include <clib/debug_protos.h>
#include "stackdump.h"
#include "strings.h"
#include "stdio.h"
#include "../../../lib/libdebugstack/libdebugstack.h"

#define DEBUG_SHOWHISTORY(x)    ;
#define	BUFSIZE 1024

extern struct ExecBase	*SysBase;
extern struct Library	*SysDebugBase;
extern struct DosLibrary *DOSBase;
extern UBYTE		*Emulation;
extern ULONG		EmulationSize;
extern UBYTE		*Module;
extern ULONG		ModuleSize;
extern ULONG            InternalFlag;

ULONG	CheckValidPtr(UBYTE	*Ptr)
{
  if (InternalFlag)
  {
    if ((Ptr >= Emulation) && (Ptr < (Emulation + EmulationSize)))
    {
      return(TRUE);
    }
    else
    if ((Ptr >= Module) && (Ptr < (Module + ModuleSize)))
    {
      return(TRUE);
    }
  }
  return(TypeOfMem(Ptr));
}

ULONG	ShowPtr(char *Name,
		UBYTE *Ptr)
{
char	*MyName;
ULONG	Hunk;
ULONG	Offset;

  if (((Ptr >= Emulation) && (Ptr < (Emulation + EmulationSize))) ||
      ((Ptr >= Module) && (Ptr < (Module + ModuleSize))) ||
      TypeOfMem(Ptr))
  {

   /*
    * would need a forbid here to protect the name
    */
    if ((MyName=SysDebugFindSeg(Ptr,
                                &Hunk,
                                &Offset)))
    {
      Printf("%s -> Address 0x%08lx -> %s Hunk %ld Offset 0x%08lx\n",
             Name,
             Ptr,
             MyName,
             Hunk,
             Offset);
      return(TRUE);
    }
    else
    {
      if (InternalFlag)
      {
        if ((Ptr >= Emulation) && (Ptr < (Emulation + EmulationSize)))
        {
          Printf("%s -> Address 0x%08lx -> ABOX: Emulation Offset 0x%lx\n",
                 Name,
                 Ptr,
                 (UBYTE*) Ptr - Module);
          return(TRUE);
        }
        else
        if ((Ptr >= Module) && (Ptr < (Module + ModuleSize)))
        {
          Printf("%s -> Address 0x%08lx -> ABOX: Module Offset 0x%lx\n",
                 Name,
                 Ptr,
                 (UBYTE*) Ptr - Module);

          return(TRUE);
        }
      }
    }
  }
  else
  {
#if 0
    Printf("Address 0x%08lx\n",
           Ptr);
#endif
  }
  return(FALSE);
}

static char *GetTaskName(struct Task *MyTask)
{
static char Buffer[512];

  if (MyTask->tc_Node.ln_Type == NT_PROCESS)
  {
    struct Process *me = (struct Process *) MyTask;
    struct CommandLineInterface *CLI = BADDR(me->pr_CLI);
    CONST_STRPTR cmdname;

    if (CLI && (cmdname = BADDR(CLI->cli_CommandName)) && *cmdname)
    {
#if 0
      UBYTE	*str;
      ULONG	len;
      str = BADDR(CLI->cli_CommandName);
      //len = min(str[0],BufLen);
      len = str[0];
      strncpy(Buffer,&str[1],len);
      Buffer[len] = '\0';
#endif

      //snprintf(Buffer,sizeof(Buffer)-1,"0x%lx <%s> [ %b ]",MyTask,MyTask->tc_Node.ln_Name,CLI->cli_CommandName);
      NewRawDoFmt("0x%lx <%s> [ %b ]",RAWFMTFUNC_STRING,Buffer,MyTask,MyTask->tc_Node.ln_Name,CLI->cli_CommandName);
    }
    else
    {
      snprintf(Buffer,sizeof(Buffer)-1,"0x%lx <%s>",MyTask,MyTask->tc_Node.ln_Name);
    }
  }
  else
  {
    snprintf(Buffer,sizeof(Buffer)-1,"0x%lx <%s>",MyTask,MyTask->tc_Node.ln_Name);
  }
  return(Buffer);
}

static ULONG ScanExecList(struct List *MyList,
                          UBYTE *Ptr,
                          char	*Buffer)
{
struct Node	*MyNode;
ULONG		Status=FALSE;

  Forbid();
  if ((MyNode=MyList->lh_Head))
  {
    while (MyNode->ln_Succ)
    {
      if (MyNode == (struct Node*) Ptr)
      {
        snprintf(Buffer,BUFSIZE-1,"<%s>",MyNode->ln_Name);
        Status = TRUE;
        break;
      }
      else
      {
        MyNode=MyNode->ln_Succ;
      }
    }
  }
  Permit();
  return(Status);
}

static ULONG ScanExecPorts(UBYTE *Ptr,
                           char	*Buffer,
                           ULONG BufferSize)
{
struct Node	*MyNode;
ULONG		Status=FALSE;

  Buffer[BufferSize-1] = '\0';
  Forbid();
  if ((MyNode=SysBase->PortList.lh_Head))
  {
    while (MyNode->ln_Succ)
    {
      if (MyNode == (struct Node*) Ptr)
      {
        snprintf(Buffer,BufferSize-1,"<%s> Task %s",MyNode->ln_Name,GetTaskName(((struct MsgPort*)MyNode)->mp_SigTask));
        Status = TRUE;
        break;
      }
      else
      {
        MyNode=MyNode->ln_Succ;
      }
    }
  }
  Permit();
  return(Status);
}

static ULONG ScanExecSemaphores(UBYTE *Ptr,
                                char	*Buffer,
                                ULONG    BufferSize)
{
struct Node	*MyNode;
ULONG		Status=FALSE;

  Buffer[BufferSize-1] = '\0';
  Forbid();
  if ((MyNode=SysBase->SemaphoreList.lh_Head))
  {
    while (MyNode->ln_Succ)
    {
      if (MyNode == (struct Node*) Ptr)
      {
        if (((struct SignalSemaphore*) MyNode)->ss_Owner)
	{
          snprintf(Buffer,BufferSize-1,"<%s> Owner %s",MyNode->ln_Name,GetTaskName(((struct SignalSemaphore*) MyNode)->ss_Owner));
	}
	else
	{
          snprintf(Buffer,BufferSize-1,"<%s> NoOwner",MyNode->ln_Name);
	}
        Status = TRUE;
        break;
      }
      else
      {
        MyNode=MyNode->ln_Succ;
      }
    }
  }
  Permit();
  return(Status);
}

static ULONG ScanExecLibs(struct List	*MyList,
                          UBYTE		*Ptr,
                          char		*Buffer)
{
struct Node	*MyNode;
ULONG		Status=FALSE;
  Forbid();
  if ((MyNode=MyList->lh_Head))
  {
    while (MyNode->ln_Succ)
    {
      if ((Ptr < (UBYTE*) MyNode) &&
          (Ptr > (UBYTE*) MyNode - ((struct Library*) MyNode)->lib_NegSize))
      {
        ULONG Offset;
        Offset = (UBYTE*) MyNode - Ptr;
        snprintf(Buffer,BUFSIZE-1,"<%s> FuncTable Offset -0x%lx",MyNode->ln_Name,Offset);
        Status = TRUE;
        break;
      }
      else
      if ((Ptr >= (UBYTE*) MyNode) &&
          (Ptr < (UBYTE*) MyNode + ((struct Library*) MyNode)->lib_PosSize))
      {
        ULONG Offset;
        Offset = (UBYTE*) Ptr - (UBYTE*) MyNode;
        snprintf(Buffer,BUFSIZE-1,"<%s> Base Offset 0x%lx",MyNode->ln_Name,Offset);
        Status = TRUE;
        break;
      }
      else
      {
        MyNode=MyNode->ln_Succ;
      }
    }
  }
  Permit();
  return(Status);
}


static ULONG CheckTask(struct Task	*MyTask,
                      UBYTE		*Ptr,
                      char		*Buffer,
                      ULONG              BufferSize)
{
  Buffer[BufferSize-1] = '\0';
  if ((struct Task*) Ptr == MyTask)
  {
    snprintf(Buffer,BufferSize-1,"%s",GetTaskName(MyTask));

    return(TRUE);
  }
  else
  if ((Ptr > (UBYTE*) MyTask) &&
      (Ptr < ((UBYTE*) MyTask + sizeof(struct Task))))
  {
    ULONG Offset;
    Offset = (UBYTE*) Ptr - (UBYTE*) MyTask;
    snprintf(Buffer,BufferSize-1,"%s Offset 0x%lx",GetTaskName(MyTask),Offset);
    return(TRUE);
  }
  else
  if ((MyTask->tc_Node.ln_Type == NT_PROCESS) &&
      (Ptr > (UBYTE*) MyTask) &&
      (Ptr < ((UBYTE*) MyTask + sizeof(struct Process))))
  {
    ULONG Offset;

    Offset = (UBYTE*) Ptr - (UBYTE*) MyTask;

    snprintf(Buffer,BufferSize-1,"%s Offset 0x%lx",GetTaskName(MyTask),Offset);
    return(TRUE);
  }
  else
  if ((Ptr > (UBYTE*) MyTask->tc_ETask) &&
      (Ptr < ((UBYTE*) MyTask->tc_ETask + sizeof(struct ETask))))
  {
    ULONG Offset;
    Offset = (UBYTE*) Ptr - (UBYTE*) MyTask->tc_ETask;
    snprintf(Buffer,BufferSize-1,"%s ETask Offset 0x%lx",GetTaskName(MyTask),Offset);
    return(TRUE);
  }
  else
  if ((Ptr >= (UBYTE*) MyTask->tc_SPLower) &&
      (Ptr < ((UBYTE*) MyTask->tc_SPUpper)))
  {
    ULONG Offset;

    Offset = (UBYTE*) Ptr - (UBYTE*) MyTask->tc_SPLower;

    snprintf(Buffer,BufferSize-1,"%s 68kStack Offset 0x%lx [0x%lx..0x%lx]",GetTaskName(MyTask),Offset,MyTask->tc_SPLower,MyTask->tc_SPUpper);
    return(TRUE);
  }
  else
  if ((Ptr >= (UBYTE*) MyTask->tc_ETask->PPCSPLower) &&
      (Ptr < ((UBYTE*) MyTask->tc_ETask->PPCSPUpper)))
  {
    ULONG Offset;

    Offset = (UBYTE*) Ptr - (UBYTE*) MyTask->tc_ETask->PPCSPLower;

    snprintf(Buffer,BufferSize-1,"%s PPCStack Offset 0x%lx [0x%lx..0x%lx]",GetTaskName(MyTask),Offset,MyTask->tc_ETask->PPCSPLower,MyTask->tc_ETask->PPCSPUpper);
    return(TRUE);
  }
  else
  if (MyTask->tc_Node.ln_Type == NT_PROCESS)
  {
    struct Process *MyProcess;
    MyProcess =(struct Process*) MyTask;
    if (MyProcess->pr_CIS &&
        ((Ptr ==(UBYTE*) MyProcess->pr_CIS) ||
         (Ptr ==(UBYTE*)  BADDR(MyProcess->pr_CIS))))
    {
      snprintf(Buffer,BufferSize-1,"%s CIS",GetTaskName(MyTask));
      return(TRUE);
    }
    else
    if (MyProcess->pr_COS &&
        ((Ptr ==(UBYTE*)  MyProcess->pr_COS) ||
         (Ptr ==(UBYTE*)  BADDR(MyProcess->pr_COS))))
    {
      snprintf(Buffer,BufferSize-1,"%s COS",GetTaskName(MyTask));
      return(TRUE);
    }
    else
    if (MyProcess->pr_CES &&
        ((Ptr ==(UBYTE*)  MyProcess->pr_CES) ||
         (Ptr ==(UBYTE*)  BADDR(MyProcess->pr_CES))))
    {
      snprintf(Buffer,BufferSize-1,"%s CES",GetTaskName(MyTask));
      return(TRUE);
    }
    else
    if (MyProcess->pr_CurrentDir &&
        ((Ptr ==(UBYTE*)  MyProcess->pr_CurrentDir) ||
         (Ptr ==(UBYTE*)  BADDR(MyProcess->pr_CurrentDir))))
    {
      snprintf(Buffer,BufferSize-1,"%s CurrentDir Lock",GetTaskName(MyTask));
      return(TRUE);
    }
    else
    if (MyProcess->pr_CLI &&
        ((Ptr ==(UBYTE*)  MyProcess->pr_CLI) ||
         (Ptr ==(UBYTE*)  BADDR(MyProcess->pr_CLI))))
    {
      snprintf(Buffer,BufferSize-1,"%s CLI",GetTaskName(MyTask));
      return(TRUE);
    }
  }
  return(FALSE);
}

static ULONG ScanTaskList(struct List	*MyList,
                          UBYTE		*Ptr,
                          char		*Buffer,
                          ULONG          BufferSize)
{
struct Node	*MyNode;
ULONG		Status=FALSE;

  if ((MyNode=MyList->lh_Head))
  {
    while (MyNode->ln_Succ)
    {
      if ((Status=CheckTask((struct Task*) MyNode,Ptr,Buffer,BufferSize)))
      {
        break;
      }
      else
      {
        MyNode=MyNode->ln_Succ;
      }
    }
  }
  return(Status);
}

static ULONG ScanTasks(UBYTE		*Ptr,
                       char		*Buffer,
                       ULONG             BufferSize)
{
struct Node	*MyNode;
ULONG		Status=FALSE;
//char		TBuffer[64];
  Disable();
  if (!(Status=CheckTask(FindTask(NULL),Ptr,Buffer,BufferSize)))
  {
    if (!(Status=ScanTaskList(&SysBase->TaskReady,Ptr,Buffer,BufferSize)))
    {
      if (!(Status=ScanTaskList(&SysBase->TaskWait,Ptr,Buffer,BufferSize)))
      {
      }
    }
  }
  Enable();
  return(Status);
}

void	ShowReg(ULONG Reg,
		UBYTE *Ptr)
{
char Buffer[BUFSIZE];

  sprintf(Buffer,"     GPR[%02ld]",Reg);

  if (!ShowPtr(Buffer,Ptr))
  {
    BOOL r1=FALSE,r2=FALSE,r3=FALSE;
    if ((r1=ScanExecList(&SysBase->LibList, Ptr, Buffer)) ||
        (r2=ScanExecList(&SysBase->DeviceList, Ptr, Buffer)) ||
        (r3=ScanExecList(&SysBase->ResourceList, Ptr, Buffer)))
    {
      Printf("     GPR[%02ld] -> Address 0x%lx -> %s %s\n",
             Reg,
             Ptr,
             r1 ? "Library" : (r2 ? "Device" : (r3 ? "Resource" : NULL)),
             Buffer);
    }
    else
    if ((r1=ScanExecLibs(&SysBase->LibList, Ptr, Buffer)) ||
        (r2=ScanExecLibs(&SysBase->DeviceList, Ptr, Buffer)))
    {
      Printf("     GPR[%02ld] -> Address 0x%lx -> %s %s\n",
             Reg,
             Ptr,
             r1 ? "Library" : "Device",
             Buffer);
    }
    else
    if ((ScanTasks(Ptr,Buffer,sizeof(Buffer))))
    {
      Printf("     GPR[%02ld] -> Address 0x%lx -> Task %s\n",
             Reg,
             Ptr,
             Buffer);
    }
    else
    if ((ScanExecPorts(Ptr,
                       Buffer ,sizeof(Buffer))))
    {
      Printf("     GPR[%02ld] -> Address 0x%lx -> Port %s\n",
             Reg,
             Ptr,
             Buffer);
    }
    else
    if ((ScanExecSemaphores(Ptr,
                            Buffer, sizeof(Buffer))))
    {
      Printf("     GPR[%02ld] -> Address 0x%lx -> Semaphore %s\n",
             Reg,
             Ptr,
             Buffer);
    }
  }
}


void	ShowPPCStackHistory(ULONG		*Stack,
                            ULONG		*StackEnd,
                            struct SegEntry	*SegArray,
                            ULONG               MaxLevel)
{
struct PPCVarArgsCodeEntry *MyPPCVarArgsCodeEntry;
char	*MyName;
ULONG	Hunk;
ULONG	Offset;
ULONG	*CurrentStack;
ULONG	*StackPtr;
int	i;


  i		=	0;
  CurrentStack	=	Stack;

  DEBUG_SHOWHISTORY(kprintf("ShowPPCStackHistory: Stack 0x%lx StackEnd 0x%lx\n",
                            Stack,
                            StackEnd));

  while (CurrentStack < StackEnd)
  {
    if (CheckValidPtr((UBYTE*) CurrentStack))
    {
      StackPtr	=(ULONG*) CurrentStack[0];	/* Get previous stackframe */
      DEBUG_SHOWHISTORY(kprintf("ShowPPCStackHistory: StackPtr 0x%lx\n",
                                StackPtr));
      if ((StackPtr >= Stack) &&
          (StackPtr < StackEnd))
      {
        if (CheckValidPtr((UBYTE*) StackPtr[1]))
        {
          if ((MyName=SysDebugFindSeg(StackPtr[1],
                                      &Hunk,
                                      &Offset)))
          {
            SegArray->Segment	=	Hunk;
            SegArray->Offset	=	Offset;

            DEBUG_SHOWHISTORY(kprintf("StackFrame[-%ld].LR-> Address 0x%08lx -> %s Hunk %ld Offset 0x%08lx\n",
                                      (ULONG) i,
                                      StackPtr[1],
                                      MyName,
                                      Hunk,
                                      Offset));
          }
          else
          {
            SegArray->Segment	=	0;
            SegArray->Offset	=	0;
            DEBUG_SHOWHISTORY(kprintf("StackFrame[-%ld].LR-> Address 0x%08lx\n",
                                      (ULONG) i,
                                      StackPtr[1]));
          }

          SegArray->Address	=(void*) StackPtr[1];
          SegArray->Name	=	MyName;
          SegArray->Function	=	NULL;



          /* Check for special code with an unnormal StackFrame
           */
          if ((MyPPCVarArgsCodeEntry=DebugStack_FindPPCVarArgsCode(StackPtr[1])))
          {
          	DEBUG_SHOWHISTORY(kprintf("Crap Code detection\n"));
          	DEBUG_SHOWHISTORY(kprintf("<%s>\n",MyPPCVarArgsCodeEntry->Name));
		DEBUG_SHOWHISTORY(kprintf("%s: StackPtr 0x%lx\n",__func__,StackPtr));
		//DEBUG_SHOWHISTORY(kprintf("%s: [0x%08lx] 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",__func__,StackPtr-4,StackPtr[-4+0],StackPtr[-4+1],StackPtr[-4+2],StackPtr[-4+3]));
		DEBUG_SHOWHISTORY(kprintf("%s: [0x%08lx] 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",__func__,StackPtr-0,StackPtr[0],StackPtr[1],StackPtr[2],StackPtr[3]));
		DEBUG_SHOWHISTORY(kprintf("%s: [0x%08lx] 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",__func__,StackPtr-4,StackPtr[4+0],StackPtr[4+1],StackPtr[4+2],StackPtr[4+3]));
		DEBUG_SHOWHISTORY(kprintf("%s: [0x%08lx] 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",__func__,StackPtr-8,StackPtr[8+0],StackPtr[8+1],StackPtr[8+2],StackPtr[8+3]));
		DEBUG_SHOWHISTORY(kprintf("%s: [0x%08lx] 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",__func__,StackPtr-12,StackPtr[12+0],StackPtr[12+1],StackPtr[12+2],StackPtr[12+3]));
		DEBUG_SHOWHISTORY(kprintf("%s: [0x%08lx] 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",__func__,StackPtr-16,StackPtr[16+0],StackPtr[16+1],StackPtr[16+2],StackPtr[16+3]));
		DEBUG_SHOWHISTORY(kprintf("%s: [0x%08lx] 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",__func__,StackPtr-20,StackPtr[20+0],StackPtr[20+1],StackPtr[20+2],StackPtr[20+3]));
		DEBUG_SHOWHISTORY(kprintf("%s: [0x%08lx] 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",__func__,StackPtr-24,StackPtr[24+0],StackPtr[24+1],StackPtr[24+2],StackPtr[24+3]));
		SegArray->Function	=	MyPPCVarArgsCodeEntry->Name;

		/*
		 * Skip the wrong Frame
		 */
		StackPtr =(ULONG*) ((UBYTE*) StackPtr + MyPPCVarArgsCodeEntry->StackOffset);
          }

          SegArray++;
          i++;
#if 0
          PPCDissasemble((APTR) StackPtr[1],
                         4,
                         PPCLibBase);
#endif
        }
        else
        {
          DEBUG_SHOWHISTORY(kprintf("StackFrame[-%ld].LR-> Address 0x%08lx **Not Valid**\n",
                  (ULONG) i,
                  StackPtr[1]));
//          break;
        }

        CurrentStack	=	StackPtr;
        if (i > MaxLevel)
        {
          DEBUG_SHOWHISTORY(kprintf("ShowPPCStackHistory: MaxLevel end\n"));
          break;
        }
      }
      else
      {
        DEBUG_SHOWHISTORY(kprintf("ShowPPCStackHistory: address outside\n"));
        break;
      }
    }
    else
    {
      DEBUG_SHOWHISTORY(kprintf("ShowPPCStackHistory: illegal StackFrame ptr\n"));
      break;
    }
  }
}

