;/*
sc saveds nostackcheck nostartup link debug line Type.c
quit
 
    (C) 1995-2001 AROS - The Amiga Research OS
    (C) 2002-2003 Harry Sintonen
    $Id: type.c,v 1.3 2022/02/15 18:40:11 piru Exp $
 
    Desc: Type CLI command
    Lang: English
*/
#include <exec/devices.h>
#include <exec/io.h>
#include <exec/memory.h>
#include <exec/execbase.h>
#include <exec/semaphores.h>
#include <dos/dos.h>
#include <dos/exall.h>
#include <proto/exec.h>
#include <proto/dos.h>


#ifdef __SASC
typedef ULONG IPTR;
#endif /* __SASC */


#include "type_version.h"

static const char version[] = VERSTAG;

#define BUFSIZE		8192
#define MAX_PATH_LEN	512

#define TEMPLATE "FROM/A/M,TO/K,OPT/K,HEX/S,NUMBER/S,NOLINE/S"
enum
{
    ARG_FROM = 0,
    ARG_TO,
    ARG_OPT,
    ARG_HEX,
    ARG_NUMBER,
    ARG_NOLINE,
    NUMARGS
};

struct file
{
	BPTR fd;
	UBYTE *cur;
	ULONG cnt;
	LONG error;
	UBYTE buf[BUFSIZE];
};

struct data
{
	struct file		in;
	struct file		out;

#ifdef __SASC
	struct AnchorPath __aligned	apath;
#else
struct AnchorPath		apath __attribute__ ((aligned (4)));
#endif
	UBYTE				pathbuf[MAX_PATH_LEN - 1];

	struct ExecBase		*SysBase;
	struct DosLibrary	*DOSBase;
};


static const UBYTE hs[] = "0123456789ABCDEF";

static
LONG processfile(CONST_STRPTR name, struct data *data, IPTR *args, LONG *numfiles);

static
LONG hexdumpfile(struct data *data);

static
LONG dumpfile(struct data *data, LONG showline, LONG noline);

int main (void)
{
	struct ExecBase	*SysBase;
	struct DosLibrary	*DOSBase;
	int retval = RETURN_FAIL;

	SysBase = *((struct ExecBase**) 4);

	DOSBase = (struct DosLibrary *)OpenLibrary("dos.library",37);
	if (DOSBase)
	{
		IPTR args[NUMARGS] = { 0 };
		struct RDArgs *rda;

		rda = ReadArgs(TEMPLATE, (LONG *) args, NULL);
		if (rda)
		{
			struct data *data;
			const UBYTE *str;

			/* Kick 1.x backwards compatibility: Handle OPT H and N
			*/
			str = (const UBYTE *) args[ARG_OPT];
			while (str)
			{
				switch (*str++)
				{
				case '\0':
					str = NULL;
					break;
				case 'h':
				case 'H':
					args[ARG_HEX] = DOSTRUE;
					break;
				case 'n':
				case 'N':
					args[ARG_NUMBER] = DOSTRUE;
					break;
				default:
					Printf("Option '%lc' ignored\n", str[-1]);
				}
			}

			/* Can't have both HEX and NUMBER
			*/
			if (args[ARG_HEX] && args[ARG_NUMBER])
			{
				FreeArgs(rda);

				PutStr("Type can't do both HEX and NUMBER\n");
				CloseLibrary((struct Library *) DOSBase);

				return RETURN_WARN;
			}

			data  = AllocMem(sizeof(struct data), MEMF_PUBLIC);
			if (data)
			{
				data->SysBase = SysBase;
				data->DOSBase = DOSBase;

				/* Pick a destination file
				*/
				if (args[ARG_TO])
					data->out.fd = Open((STRPTR) args[ARG_TO], MODE_NEWFILE);
				else
					data->out.fd = Output();

				if (data->out.fd)
				{
					STRPTR *names;

					data->out.cur = data->out.buf;
					data->out.cnt = BUFSIZE;

					data->apath.ap_BreakBits  = SIGBREAKF_CTRL_C;
					data->apath.ap_FoundBreak = 0;
					data->apath.ap_Flags      = 0;
					data->apath.ap_Strlen     = APSF_EXTENDED | MAX_PATH_LEN;
					data->apath.ap_Extended   = APEF_LiteralSLinks;

					for (names = (STRPTR *) args[ARG_FROM]; *names; names++)
					{
						ULONG numfiles = 0;
						LONG error;

						for (error = MatchFirst(*names, &data->apath);
						     !error;
						     error = MatchNext(&data->apath))
						{
							error = processfile(data->apath.ap_Buf, data, args, &numfiles);
							if (error)
								break;
						}
						MatchEnd(&data->apath);

						if (numfiles == 0 && error == ERROR_NO_MORE_ENTRIES)
						{
							error = -1;
						}

						if (error && error != ERROR_NO_MORE_ENTRIES)
						{
							if (*names && error != ERROR_BREAK)
							{
								Printf("TYPE: can't open %s\n", (LONG) *names);
							}

							if (error != -1)
							{
								PrintFault(error, NULL);
								SetIoErr(error);
							}

							break;
						}
					}

					if (args[ARG_TO])
						Close(data->out.fd);

					/* If all files got dumped, return ok, else error.
					*/
					retval = *names ? RETURN_ERROR : RETURN_OK;
				}
				else
				{
					PrintFault(IoErr(), NULL);
				}

				FreeMem(data, sizeof(struct data));
			}
			else
				PrintFault(ERROR_NO_FREE_STORE, NULL);

			FreeArgs(rda);
		}
		else
		{
			PrintFault(IoErr(), NULL);
		}

		CloseLibrary((struct Library *)DOSBase);
	}

	return retval;
}

#define SysBase data->SysBase
#define DOSBase data->DOSBase

static
LONG processfile(CONST_STRPTR name, struct data *data, IPTR *args, LONG *numfiles)
{
	LONG error = 0;

	data->in.fd = Open(name, MODE_OLDFILE);
	if (data->in.fd)
	{
		(*numfiles)++;
		data->in.cnt = 0;

		if (args[ARG_HEX])
			error = hexdumpfile(data);
		else
			error = dumpfile(data, args[ARG_NUMBER], args[ARG_NOLINE]);

		Close(data->in.fd);
	}
	else
		error = IoErr();

	return error;
}


#define putc(f,c) (*(f)->cur++ = (c), --(f)->cnt ? 0 : put(f, data))

static
int put(struct file *f, struct data *data)
{
	LONG size,subsize;
	STRPTR buf;

	size = f->cur-f->buf;
	buf = f->buf;

	while (size)
	{
		subsize = Write(f->fd,buf,size);
		if (subsize <= 0)
		{
			f->error = IoErr();

			return 1;
		}
		buf  += subsize;
		size -= subsize;
	}
	f->cur = f->buf;
	f->cnt = BUFSIZE;

	return 0;
}


#define getc(f) ((f)->cnt-- ? *(f)->cur++ : get(f, data))

static
int get(struct file *f, struct data *data)
{
	LONG size;

	size = Read(f->fd, f->buf, BUFSIZE);

	if (size < 0)
		f->error = IoErr();

	if (size <= 0)
		return -1;

	f->cnt = size - 1;
	f->cur = f->buf;

	return *f->cur++;
}


static
void putlinequick(struct file *f, ULONG offset, UBYTE *buf)
{
	int c, i, k;
	UBYTE *b, *o;

	o = f->cur;

	if (offset >= 0x10000)
	{
		if (offset >= 0x100000)
		{
			if (offset >= 0x1000000)
			{
				if (offset >= 0x10000000)
				{
					*o++ = hs[(offset >> 28) & 0xf];
					f->cnt--;
				}
				*o++=hs[(offset >> 24) &0xf];
				f->cnt--;
			}
			*o++=hs[(offset >> 20) & 0xf];
			f->cnt--;
		}
		*o++=hs[(offset >> 16) & 0xf];
		f->cnt--;
	}
	*o++ = hs[(offset >> 12) & 0xf];
	*o++ = hs[(offset >> 8) & 0xf];
	*o++ = hs[(offset >> 4) & 0xf];
	*o++ = hs[offset & 0xf];
	*o++ = ':';
	*o++ = ' ';

	b=buf;
	for (i = 0; i < 4; i++)
	{
		for (k = 0; k < 4; k++)
		{
			c = *b++;
			*o++ = hs[c >> 4];
			*o++ = hs[c & 0xf];
		}
		*o++=' ';
	}
	b=buf;
	for (i = 0; i < 16; i++)
	{
		c = *b++;
		*o++ = (c & 0x7f) >= 0x20 && c != 0x7f ? c : '.';
	}
	*o++ = '\n';
	f->cur = o;
	f->cnt -= 59;
}


static
int putline(struct file *f, ULONG offset, UBYTE *buf, ULONG num, struct data *data)
{
	int c, i;
	UBYTE *b;

	if (offset >= 0x10000)
	{
		if (offset >= 0x10000000 && putc(f, hs[(offset >> 28) & 0xf]))
			return 1;

		if (offset >= 0x1000000 && putc(f, hs[(offset >> 24) & 0xf]))
			return 1;

		if (offset >= 0x100000 && putc(f, hs[(offset >> 20) & 0xf]))
			return 1;

		if (offset >= 0x10000 && putc(f, hs[(offset >> 16) & 0xf]))
			return 1;
	}

	if (putc(f, hs[(offset >> 12) & 0xf]))
		return 1;

	if (putc(f, hs[(offset >> 8) & 0xf]))
		return 1;

	if (putc(f, hs[(offset >> 4) & 0xf]))
		return 1;

	if (putc(f, hs[offset & 0xf]))
		return 1;

	if (putc(f, ':'))
		return 1;

	if (putc(f, ' '))
		return 1;

	b=buf;
	for (i = 0;i < 16; i++)
	{
		if (i < num)
		{
			c = *b++;
			if (putc(f, hs[c >> 4]))
				return 1;

			if (putc(f, hs[c & 0xf]))
				return 1;
		}
		else
		{
			if (putc(f, ' '))
				return 1;

			if (putc(f, ' '))
				return 1;
		}

		if ((i & 3) == 3 && putc(f,' '))
			return 1;
	}

	b=buf;
	for (i = 0; i < num; i++)
	{
		c=*b++;
		if (putc(f,(c & 0x7f) >= 0x20 && c != 0x7f ? c : '.'))
			return 1;
	}

	if (putc(f, '\n'))
		return 1;

	return 0;
}


static
LONG hexdumpfile(struct data *data)
{
	struct file *in = &data->in;
	struct file *out = &data->out;
	UBYTE buf[16];
	UBYTE *b;
	LONG offset=0, n, c;
	int count=0;
	//LONG tty;

	//tty = IsInteractive(out->fd);

	in->error  = 0;
	out->error = 0;

	for(;;)
	{
		if (((++count & 0xf) == 0) && (SetSignal(0, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C))
		{
			return ERROR_BREAK;
		}

		if (in->cnt > 16)
		{
			b = in->cur;
			n = 16;
			in->cur += 16;
			in->cnt -= 16;
		}
		else
		{
			b = buf;
			for(n = 0; n < 16; n++)
			{
				c = getc(in);
				if (c < 0)
					break;

				b[n] = c;
			}
		}

		if (n == 16)
		{
			if (out->cnt >= 63)
				putlinequick(out, offset, b);
			else if (putline(out, offset, b, n, data))
				return out->error;
		}
		else
		{
			if (n)
			{
				if (putline(out, offset, b, n, data))
					return out->error;

				if (putc(out, '\n'))
					return out->error;
			}

			if (out->cur != out->buf && put(out, data))
				return out->error;

			return in->error;
		}

		//if (tty && put(out))
		//    return out->error;
		if (put(out, data))
			return out->error;

		offset += n;
	}
}


static
int putlinenumber(struct file *out, unsigned short line, struct data *data)
{
	int x = 10000;
	BOOL s = FALSE;

	while (x)
	{
		if (line / x || s)
		{
			if (putc(out, line / x + '0'))
				return 1;

			line %= x;
			s = TRUE;
		}
		else if putc(out, ' ')
			return 1;

		x /= 10;
	}

	putc(out, ' ');

	return 0;
}


static
LONG dumpfile(struct data *data, LONG showline, LONG noline)
{
	struct file *in = &data->in;
	struct file *out = &data->out;
	LONG c, lastc = 0;
	unsigned short line = 0;
	int count = 0;

	in->error  = 0;
	out->error = 0;

	if (showline && putlinenumber(out, ++line, data))
		return out->error;

	if (1/*IsInteractive(out->fd)*/)
	{
		for (;;)
		{
			if (((++count & 0xff) == 0) && (SetSignal(0, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C))
			{
				return ERROR_BREAK;
			}

			c = getc(in);

			if (c < 0)
			{
				if (lastc != '\n' && !noline && putc(out, '\n'))
					return out->error;

				if (put(out, data))
					return out->error;

				return in->error;
			}

			if (lastc == '\n' && showline && putlinenumber(out, ++line, data))
				return out->error;

			if (putc(out,c) || (c == '\n' && put(out, data)))
				return out->error;

			lastc = c;
		}
	}
}

